import os
import shutil
import requests
import unittest
from orderedattrdict import AttrDict
from . import server
from gramex import conf
from gramex.http import OK
from gramex.services import info

tempfiles = AttrDict()
folder = os.path.dirname(os.path.abspath(__file__))


def setUp():
    # Remove uploads folder before Gramex locks .meta.h5
    # This may fail on Python 2.7 on Windows due to unicode characters.
    # Delete tests/uploads/ manually in that case
    upload_path = os.path.join(folder, 'uploads')
    if os.path.exists(upload_path):
        shutil.rmtree(upload_path)
    server.start_gramex()


def tearDown():
    server.stop_gramex()

    for filename in tempfiles.values():
        if os.path.exists(filename):
            os.unlink(filename)


class TestGramex(unittest.TestCase):
    '''Base class to test Gramex running as a subprocess'''

    def get(self, url, session=None, method='get', timeout=10, **kwargs):
        req = session or requests
        return getattr(req, method)(server.base_url + url, timeout=timeout, **kwargs)

    def check(self, url, data=None, path=None, code=200, text=None, no_text=None,
              request_headers=None, headers=None, session=None, method='get', timeout=10):
        '''
        check(url) checks if the url returns the correct response. Parameters:

        :arg string url: Relative URL from test server base
        :arg dict data: optional data= to pass to requests.get/post
        :arg dict request_headers: options headers= to pass to requests.get/post
        :arg string method: HTTP method (default: 'get', may be 'post', etc.)
        :arg Session session: requests.Session object to use (default: ``requests``)
        :arg int timeout: seconds to wait (default: 10)

        :arg int code: returned status code must match this (default: 200)
        :arg string text: returned body must contain this text
        :arg string no_text: returned body must NOT contain this text
        :arg string path: returned body must equal the contents of the file at this path
        :arg dict headers: returned headers must contain these items. Value can be:
            - None/False: the header SHOULD NOT exist
            - True: the header SHOULD exist
            - string: the header must equal this string

        If any of the checks do not match, raises an assertion error.
        '''
        r = self.get(url, session=session, data=data, method=method, timeout=timeout,
                     headers=request_headers)
        self.assertEqual(r.status_code, code, '%s: code %d != %d' % (url, r.status_code, code))
        if text is not None:
            self.assertIn(text, r.text, '%s: %s not in %s' % (url, text, r.text))
        if no_text is not None:
            self.assertNotIn(text, r.text, '%s: %s should not be in %s' % (url, text, r.text))
        if path is not None:
            with (server.info.folder / path).open('rb') as file:
                self.assertEqual(r.content, file.read(), '%s != %s' % (url, path))
        if headers is not None:
            for header, value in headers.items():
                if value is None or value is False:
                    self.assertFalse(header in r.headers,
                                     '%s: should not have header %s' % (url, header))
                elif value is True:
                    self.assertTrue(header in r.headers,
                                    '%s: should have header %s' % (url, header))
                else:
                    actual = r.headers[header]
                    self.assertEqual(actual, value,
                                     '%s: header %s = %s != %s' % (url, header, actual, value))
        return r


class TestSchedule(TestGramex):

    def test_schedule(self):
        # Run this test as soon as Gramex starts to check if schedule has run.
        self.assertIn('schedule-key', info, 'Schedule was run at startup')
        self.check('/', code=OK)

        # This tests that long running threads run in parallel. We run
        # utils.slow_count every 10ms for kwargs.count seconds. If this test
        # fails, increase schedule.schedule-startup-slow.kwargs.count. It may be
        # due to slow startup.
        max_count = conf.schedule['schedule-startup-slow'].kwargs.count - 1
        self.assertTrue(0 < info['schedule-count'] < max_count, 'Schedule still running')



import os
import unittest
import gramex.ml
import pandas as pd
from sklearn.svm import SVC
from sklearn.naive_bayes import BernoulliNB
from gramex.ml import Classifier

from nose.tools import eq_, ok_
from . import folder


class TestClassifier(unittest.TestCase):
    model_path = os.path.join(folder, 'model.pkl')

    def test_01_train(self):
        path = os.path.join(folder, 'iris.csv')
        data = pd.read_csv(path, encoding='utf-8')

        # Model can be trained without specifying a model class, input or output
        model1 = Classifier()
        model1.train(data)
        ok_(isinstance(model1.model, BernoulliNB))
        eq_(model1.input, data.columns[:4].tolist())
        eq_(model1.output, data.columns[-1])

        # Train accepts explicit model_class, model_kwargs, input and output
        model2 = Classifier()
        inputs = ['petal_length', 'petal_width', 'sepal_length', 'sepal_width']
        model2.train(
            data,                                 # DataFrame with input & output columns
            model_class='sklearn.svm.SVC',        # Any sklearn model works
            model_kwargs={'kernel': 'sigmoid'},   # Optional model parameters
            input=inputs,
            output='species'
        )
        eq_(model2.input, inputs)
        eq_(model2.output, model1.output)
        ok_(isinstance(model2.model, SVC))
        eq_(model2.model.kernel, 'sigmoid')

        # Test predictions. Note: this is manually crafted. If it fails, change the test case
        result = model1.predict([
            {'sepal_length': 5, 'sepal_width': 3, 'petal_length': 1.5, 'petal_width': 0},
            {'sepal_length': 5, 'sepal_width': 2, 'petal_length': 5.0, 'petal_width': 1},
            {'sepal_length': 6, 'sepal_width': 3, 'petal_length': 4.8, 'petal_width': 2},
        ])
        eq_(result.tolist(), ['setosa', 'versicolor', 'virginica'])

        # Test saving
        expected = model1.predict(data[model1.input])
        if os.path.exists(self.model_path):
            os.remove(self.model_path)
        model1.save(self.model_path)
        actuals = gramex.ml.load(self.model_path).predict(data[model1.input])
        eq_(actuals.tolist(), expected.tolist())

    def test_02_load(self):
        model = gramex.ml.load(self.model_path)
        result = model.predict([{
            'sepal_length': 5.7,
            'sepal_width': 4.4,
            'petal_length': 1.5,
            'petal_width': 0.4,
        }])
        eq_(result.tolist(), ['setosa'])

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(cls.model_path):
            os.remove(cls.model_path)



import os
import pandas as pd
from gramex.ml import Classifier
from nose.tools import eq_, ok_
from . import folder, TestGramex


class TestModelHandler(TestGramex):
    @classmethod
    def setUpClass(cls):
        data = pd.read_csv(os.path.join(folder, '..', 'testlib', 'iris.csv'), encoding='utf-8')
        model = Classifier()
        model.train(data)
        model.save(os.path.join(folder, 'gen.iris.pkl'))

    def test_model_info(self):
        r = self.check('/model/iris')
        info = r.json()
        ok_('BernoulliNB' in info['model'])
        eq_(info['input'], ['sepal_length', 'sepal_width', 'petal_length', 'petal_width'])
        eq_(info['output'], 'species')

    def test_predict(self):
        # Test individual results
        data = pd.DataFrame([
            {'sepal_length': 5, 'sepal_width': 3, 'petal_length': 1.5, 'petal_width': 0},
            {'sepal_length': 5, 'sepal_width': 2, 'petal_length': 5.0, 'petal_width': 1},
            {'sepal_length': 6, 'sepal_width': 3, 'petal_length': 4.8, 'petal_width': 2},
        ])
        single = [
            self.check('/model/iris', data=row.to_dict()).json()['result'][0]
            for index, row in data.iterrows()
        ]
        ok_(all(species in {'setosa', 'versicolor', 'virginica'} for species in single))

        # Test multiple results
        multi = self.check('/model/iris', data=data.to_dict(orient='list')).json()['result']
        eq_(multi, single)

        # TODO: check that the results are correct. Today, this is not working
        # eq_(multi, ['setosa', 'virginica', 'versicolor'])

    def test_predict_incomplete(self):
        self.check('/model/iris', data={'sepal_length': 5}, code=500)



import os
import inspect
import pandas as pd
from gramex.config import locate, app_log
from sklearn.externals import joblib
from sklearn.preprocessing import StandardScaler

# Expose joblob.load via gramex.ml
load = joblib.load                      # noqa


class Classifier(object):
    '''
    TODO
    '''
    def train(self, data, model_class='sklearn.naive_bayes.BernoulliNB', model_kwargs={},
              output=None, input=None, labels=None):
        '''
        :arg data DataFrame: data to train / re-train the model with
        :arg model_class str: model class to use (default: ``sklearn.naive_bayes.BernoulliNB``)
        :arg model_kwargs dict: kwargs to pass to model class constructor (defaults: ``{}``)
        :arg output str: output column name (default: last column in training data)
        :arg input list: input column names (default: all columns except ``output``)
        :arg labels list: list of possible output values (default: unique ``output`` in training)

        Notes:
        - If model has already been trained, extend the model. Else create it
        '''
        self.output = output if output is not None else data.columns[-1]
        self.input = input
        if self.input is None:
            self.input = list(data.columns)
            self.input.remove(self.output)

        # If model does not exist, create model
        if not hasattr(self, 'model'):
            # Split it into input (x) and output (y)
            x, y = data[self.input], data[self.output]

            # Transform the data
            self.scaler = StandardScaler()
            self.scaler.fit(x)

            # Train the classifier. Partially, if possible
            clf = locate(model_class)(**model_kwargs)
            if labels is not None and hasattr(clf, 'partial_fit'):
                clf.partial_fit(self.scaler.transform(x), y, classes=labels)
            else:
                clf.fit(self.scaler.transform(x), y)
            self.model = clf

        # Extend the model
        else:
            x, y = data[self.input], data[self.output]
            classes = set(self.model.classes_)
            classes |= set(y)
            self.model.partial_fit(self.scaler.transform(x), y)

    def predict(self, data):
        '''
        Return a Series that has the results of the classification of data
        '''
        # Convert list of lists or numpy arrays into DataFrame. Assume columns are as per input
        if not isinstance(data, pd.DataFrame):
            data = pd.DataFrame(data, columns=self.input)
        # Take only trained input columns
        data = data[self.input]
        return self.model.predict(self.scaler.transform(data))

    def save(self, path):
        '''
        Serializes the model and associated parameters
        '''
        joblib.dump(self, path, compress=9)


def _conda_r_home():
    '''
    Returns the R home directory for Conda R if it is installed. Else None.

    Typically, people install Conda AND R (in any order), and use the system R
    (rather than the conda R) by placing it before Conda in the PATH.

    But the system R does not work with Conda rpy2. So we check if Conda R
    exists and return its path, so that it can be used as R_HOME.
    '''
    try:
        from conda.base.context import context
    except ImportError:
        app_log.error('Anaconda not installed. Cannot use Anaconda R')
        return None
    r_home = os.path.normpath(os.path.join(context.root_prefix, 'lib', 'R'))
    if os.path.isdir(os.path.join(r_home, 'bin')):
        return r_home
    app_log.error('Anaconda R not installed')
    return None


def r(code=None, path=None, rel=True, conda=True, convert=True,
      repo='https://cran.microsoft.com/', **kwargs):
    '''
    Runs the R script and returns the result.

    :arg str code: R code to execute.
    :arg str path: R script path. Cannot be used if code is specified
    :arg bool rel: True treats path as relative to the caller function's file
    :arg bool conda: True overrides R_HOME to use the Conda R
    :arg bool convert: True converts R objects to Pandas and vice versa
    :arg str repo: CRAN repo URL

    All other keyword arguments as passed as parameters
    '''
    # Use Conda R if possible
    if conda:
        r_home = _conda_r_home()
        if r_home:
            os.environ['R_HOME'] = r_home

    # Import the global R session
    try:
        from rpy2.robjects import r, pandas2ri, globalenv
    except ImportError:
        app_log.error('rpy2 not installed. Run "conda install rpy2"')
        raise
    except RuntimeError:
        app_log.error('Cannot find R. Set R_HOME env variable')
        raise

    # Set a repo so that install.packages() need not ask for one
    r('local({r <- getOption("repos"); r["CRAN"] <- "%s"; options(repos = r)})' % repo)

    # Activate or de-activate automatic conversion
    # https://pandas.pydata.org/pandas-docs/version/0.22.0/r_interface.html
    if convert:
        pandas2ri.activate()
    else:
        pandas2ri.deactivate()

    # Pass all other kwargs as global environment variables
    for key, val in kwargs.items():
        globalenv[key] = val

    if code and path:
        raise RuntimeError('Use r(code=) or r(path=...), not both')
    if path:
        # if rel=True, load path relative to parent directory
        if rel:
            stack = inspect.getouterframes(inspect.currentframe(), 2)
            folder = os.path.dirname(os.path.abspath(stack[1][1]))
            path = os.path.join(folder, path)
        result = r.source(path, chdir=True)
        # source() returns a withVisible: $value and $visible. Use only the first
        result = result[0]
    else:
        result = r(code)

    return result



import json
import pandas as pd
import gramex.ml
import gramex.cache
from gramex.handlers import BaseHandler


class ModelHandler(BaseHandler):
    '''
    TODO
    '''
    @classmethod
    def setup(cls, path, **kwargs):
        super(ModelHandler, cls).setup(**kwargs)
        cls.path = path

    def get(self):
        model = gramex.cache.open(self.path, gramex.ml.load)
        self.set_header('Content-Type', 'application/json; charset=utf-8')

        # If no model columns are passed, return model info
        if not any(col in self.args for col in model.input):
            self.write(json.dumps({
                'model': repr(model.model),
                'input': model.input,
                'output': model.output,
            }, indent=4))
            return

        data = pd.DataFrame(self.args)
        data = data[model.input]
        data = model.scaler.transform(data)
        result = model.predict(data).tolist()
        result = {'result': result}
        self.write(result)



"""Auth module role settings."""
import sys
from tornado.web import HTTPError
from cachetools import TTLCache
from six.moves import StringIO
import gramex
from gramex.config import app_log
from gramex.http import INTERNAL_SERVER_ERROR
from gramex.handlers import FormHandler, DBAuth

contexts = TTLCache(maxsize=100, ttl=1800)


def get_auth_conf(kwargs):
    '''
    Expects kwargs.authhandler to point to an AuthHandler key in gramex config.
    The AuthHandler must have a lookup.
    Returns the authhandler, its configuration, and the FormHandler data configuration.
    Used in AdminFormHandler *and* in index.html. So keep it as a separate function.
    '''
    if 'authhandler' not in kwargs:
        raise ValueError('Missing authhandler')
    authhandler = kwargs['authhandler']
    if authhandler not in gramex.conf.get('url', {}):
        raise ValueError('Missing url.%s (cannot find authhandler)' % authhandler)
    auth_conf = gramex.conf.url[authhandler]
    auth_kwargs = auth_conf.get('kwargs', {})
    if 'lookup' in auth_kwargs:
        data_conf = auth_kwargs['lookup'].copy()
        return authhandler, auth_conf, data_conf
    elif auth_conf.get('handler', None) == 'DBAuth':
        # For DBAuth, hoist the user.column into as the id: for the URL
        user_column = auth_kwargs.get('user', {}).get('column', 'user')
        data_conf = DBAuth.clear_special_keys(auth_kwargs.copy(), 'user',
                                              'password', 'forgot', 'signup')
        data_conf['id'] = user_column
        return authhandler, auth_conf, data_conf
    else:
        raise ValueError('Missing lookup: in url.%s (authhandler)' % authhandler)


class AdminFormHandler(FormHandler):
    '''
    A customized FormHandler. Specify a "kwargs.admin_kwargs.authhandler: auth-handler".
    It lookup up "auth-handler" in the gramex config. If it has a "lookup:" or is a "DBAuth",
    creates a FormHandler using that url: and other parameters.
    '''
    @classmethod
    def setup(cls, **kwargs):
        # admin_kwargs.authhandler is a url: key that holds an AuthHandler. Get its kwargs
        try:
            authhandler, auth_conf, data_conf = get_auth_conf(kwargs.get('admin_kwargs', {}))
        except ValueError as e:
            super(FormHandler, cls).setup(**kwargs)
            app_log.warning('%s: %s', cls.name, e.args[0])
            cls.reason = e.args[0]
            cls.get = cls.post = cls.put = cls.delete = cls.send_response
            return
        # Get the FormHandler configuration from lookup:
        cls.conf.kwargs = data_conf
        super(AdminFormHandler, cls).setup(**cls.conf.kwargs)

    def send_response(self, *args, **kwargs):
        raise HTTPError(INTERNAL_SERVER_ERROR, reason=self.reason)


def evaluate(handler, code):
    """Evaluates Python code in a WebSocketHandler, like a REPL"""
    retval = None
    # Check if code is an expression (eval) or statement (exec)
    try:
        co, mode = compile(code, '<input>', 'eval'), 'eval'
    except SyntaxError:
        try:
            co, mode = compile(code, '<input>', 'exec'), 'exec'
        except Exception as e:
            retval = e
    except Exception as e:
        retval = e
    if retval is not None:
        handler.write_message(repr(retval))
        return

    # Capture stdout
    old_stdout, out = sys.stdout, StringIO()
    sys.stdout = out
    # Run code and get the result. (Result is None for exec)
    try:
        context = contexts.setdefault(handler.session['id'], {})
        if mode == 'eval':
            result = eval(co, context)
        else:
            exec(co, context)
            result = None
    except Exception as e:
        result = e
    finally:
        sys.stdout = old_stdout

    # Write the stdout (if any), then the returned value (if any)
    retval = out.getvalue()
    if result is not None:
        retval += repr(result)
    handler.write_message(retval)



'''Functions to transform data'''

from .transforms import build_transform, condition, flattener, once
from .badgerfish import badgerfish
from .template import template
from .auth import ensure_single_session
from .twitterstream import TwitterStream

__all__ = [
    'build_transform',
    'badgerfish',
    'template',
    'ensure_single_session',
    'condition',
    'flattener',
    'once',
    'TwitterStream',
]



'''
This module implements different strategies to uniquely identify machines.

http://serialsense.com/blog/2011/02/generating-unique-machine-ids/
http://stackoverflow.com/questions/2461141/get-a-unique-computer-id-in-python-on-windows-and-linux

'''

import uuid


def mac_id():
    return uuid.getnode()


def cpu_id():
    raise NotImplementedError('CPU ID not yet implemented')


def bios_id():
    raise NotImplementedError('BIOS ID not yet implemented')
